#include "CheckersAI.h"

CheckerBoard CheckersAI::findBestMove(const CheckerBoard &board, int depth)
{
    // Alpha is smallest possible number
    int alpha = std::numeric_limits<int>::min();
    // Beta is highest possible number
    int beta = std::numeric_limits<int>::max();

    // Best value is the biggest for white and the smallest for black
    int bestValue = isWhite ? std::numeric_limits<int>::min() : std::numeric_limits<int>::max();
    CheckerBoard bestMove;

    // Generate all possible moves of all checkers for the current side
    std::vector<CheckerBoard> moves = board.generateLegalMoves(isWhite);

    // Throw exception if there are no possible moves (change this to forfeit later)
    if (moves.size() == 0)
        throw exception();

    // Iterate over all possible moves
    for (const auto &move : moves)
    {
        // Calculate the minimax score for this move
        int value = minimax(move, depth - 1, alpha, beta, !isWhite);

        // If the value is better than current best
        if (isWhite && value > bestValue)
        {
            bestValue = value;                  // Save the value
            bestMove = move;                    // Save the move
            alpha = std::max(alpha, bestValue); // Update the alpha
        }
        // The best value is the smallest value for black
        else if (!isWhite && value < bestValue)
        {
            bestValue = value;                // Save the value
            bestMove = move;                  // Save the move
            beta = std::min(beta, bestValue); // Update the beta
        }
        // Alpha-beta pruning
        if (alpha >= beta)
            break;
    }
    wcout << "Best move: " << bestMove.lastMove.c_str() << endl;

    return bestMove;
}

int CheckersAI::minimax(CheckerBoard board, int depth, int alpha, int beta, bool isMaximizing)
{
    // When reached the cut-out leaf of the decision tree, evaluate the score
    if (depth == 0 || board.isGameOver())
    {
        return board.evaluate();
    }

    std::vector<CheckerBoard> moves = board.generateLegalMoves(isMaximizing);

    // Maximizing for blacks
    if (isMaximizing)
    {
        int maxEval = std::numeric_limits<int>::min();
        for (const auto &move : moves)
        {
            // Go further down the decision tree
            int eval = minimax(move, depth - 1, alpha, beta, false);
            // Keep current max evaluated score
            maxEval = std::max(maxEval, eval);
            // Update the alpha
            alpha = std::max(alpha, eval);
            // Alpha-beta pruning
            if (beta <= alpha)
                break;
        }
        // Return max evaluated score for blacks
        return maxEval;
    }
    // Minimizing for whites
    else
    {
        int minEval = std::numeric_limits<int>::max();
        for (const auto &move : moves)
        {
            // Go further down the decision tree
            int eval = minimax(move, depth - 1, alpha, beta, true);
            // Keep current max evaluated score
            minEval = std::min(minEval, eval);
            // Update ethe beta
            beta = std::min(beta, eval);
            // Alpha-beta pruning
            if (beta <= alpha)
                break;
        }
        // Return min evaluated score for whites
        return minEval;
    }
}